
package jworms.game;

import jworms.openGL.BigTexture;
import jworms.network.UDPNetworkListener;
import jworms.network.GameInfo;
import jworms.network.UDPNetwork;
import jworms.util.Queue;
import org.lwjgl.Display;
import org.lwjgl.Sys;
import org.lwjgl.input.Mouse;

import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;
import java.awt.image.PixelGrabber;
import java.awt.Point;
import java.awt.Image;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.HashMap;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.DataOutputStream;

public class GameField implements UDPNetworkListener{
// constants
  /** Constant for DotMap background */
  public static final byte DOTMAP_BACKGROUND = 0;
  /** Ground: weak */
  public static final byte DOTMAP_GROUND_WEAK = 1;
  /** Ground: hard */
  public static final byte DOTMAP_GROUND_HARD = 12;

  /** Gravity constant */
  public static final float GRAVITY = 0.1f;

  /** Next free classID. */
  private static int nextClassID = 1;
  /** HashMap from ID (Integer) to class (GameObject). */
  private static HashMap id2Class = new HashMap();

  /**
   * Registers an game object for network connection.
   *
   * @param objectClass object class
   */
  public static void registerClass(GameObject objectClass){
    objectClass.setClassID(nextClassID);
    id2Class.put(new Integer(objectClass.getClassID()), objectClass);

    System.out.println("registerClass: " + objectClass.getClassID() + " = " + objectClass);

    // check for classID errors
    for(int classID = 1; classID <= nextClassID; classID++){
      GameObject object = (GameObject) id2Class.get(new Integer(classID));
      if(object.getClassID() != classID){
        System.out.println("Error in " + object + " due to wrong classID!");
        System.exit(-1);
      }
    }

    nextClassID++;
  }

// game field properies
  /** Name of the map. */
  private String fieldName = null;

  /** Width of the landscape/ dot map. */
  private int width = 0;
  /** Height of the landscape/ dot map. */
  private int height = 0;

  /** Landscape texture foreground. */
  private BigTexture landscapeFront = null;
  /** Landscape texture background. */
  private BigTexture landscapeBack = null;
  /** Dot map. */
  private byte[] dotMap = null;

// scrolling the map
  /** Actual x,y position of the map (upper left corner). */
  private Point2D.Float position = null;

  /** Map automatically scrolls to this point (upper left corner). */
  private Point2D.Float scrollToPoint = null;
  /** Current scroll speed of the map. */
  private float scrollSpeed = 0;

  /** If an explosion lets the game field shake. */
  private float shakeStrength = 0;

  /** Time the game was started (in seconds). */
  private double startTime = 0;
  /** Time per frame (in seconds). */
  private double frameTime = 1. / 60;
  /** Current frame. */
  private long frame = 0;

  /** Objects within the game field (GameObject) */
  private ArrayList gameObjects = new ArrayList();
  /** Objects to add */
  private ArrayList gameObjectsAdd = new ArrayList();
  /** Objects to remove */
  private ArrayList gameObjectsRemove = new ArrayList();

  /** Object finder for hurting. */
  private GameFieldObjectFinder objectFinder = null;

  /** UDPNetwork. */
  private UDPNetwork network = null;
  /** Network identifier. */
  private int networkID = 0;
  /** Next free objectID. */
  private int nextObjectID = 1;
  /** HashMap of ID (Integer) to holograms (GameObject). */
  private HashMap holograms = new HashMap();

  /** Recieved packets. */
  private Queue recievedPackets = new Queue();

  /** Mouse position */
  private Point mousePosition = null;

  /**
   * Creates a game field.
   *
   * @param fieldName name of the field
   */
  public GameField(String fieldName, UDPNetwork network, int networkID) throws Exception{
    this.fieldName = fieldName;
    this.network = network;

    this.networkID = networkID;
    nextObjectID |= this.networkID << 24;

    network.addUDPNetworkListener(this);

    // load landscape
    landscapeBack = new BigTexture(this.fieldName + "_back.png", BigTexture.FORMAT_RGB, true);
    landscapeFront = new BigTexture(this.fieldName + "_front.png", BigTexture.FORMAT_RGBA, false);
    width = landscapeFront.getImage().getWidth();
    height = landscapeFront.getImage().getWidth();

    // load dot map
    Image dotMapImage = (new javax.swing.ImageIcon(this.fieldName + "_dotMap.png")).getImage();
    PixelGrabber grabber = new PixelGrabber(dotMapImage, 0, 0, width, height, false);
    try{
      grabber.grabPixels();
      dotMap = (byte[]) grabber.getPixels();
    } catch(Exception e){
      System.out.println("Loading dot map failed due to " + e);
      System.exit(-1);
    }

    // update front alpha channel
    BufferedImage landscapeFrontTexture = landscapeFront.getImage();
    for(int y = 0; y < height; y++){
      for(int x = 0; x < width; x++){
        if(getDot(x, y) == GameField.DOTMAP_BACKGROUND){
          landscapeFrontTexture.setRGB(x, y, landscapeFrontTexture.getRGB(x, y) & 0x00FFFFFF);
        }
      }
    }
    landscapeFront.update(0, 0, width, height);

    // first position = center
    position = new Point2D.Float((-width + Display.getWidth()) / 2, (-height + Display.getHeight()) / 2);
    mousePosition = new Point(width / 2, height / 2);
    scrollToPoint = position;

    // create object finder
    objectFinder = new GameFieldObjectFinder(this);
  }

  /**
   * Returns the time of the hres timer in seconds.
   *
   * @return time in seconds
   */
  private double getTime(){
    return ((double) Sys.getTime()) / Sys.getTimerResolution();
  }

  /**
   * Returns the time point for the given frame.
   *
   * @param frame frame number
   * @return time for the frame
   */
  public double getTime(long frame){
    return startTime + frame * frameTime;
  }

  /**
   * Starts the game at the given frame.
   *
   * @param frame frame
   */
  public void start(int frame){
    this.frame = frame;
    startTime = getTime() - frame * frameTime;
  }

  /**
   * Returns the number of the current frame.
   *
   * @return number of the current frame
   */
  public long getFrame(){
    return frame;
  }

  /**
   * Scrolls the game field.
   */
  public void doMove(){
    // calculate current frame
    double time = getTime();
    int currentFrame = (int) ((time - startTime) / frameTime);
    if(currentFrame == frame) currentFrame = (int) ((time - startTime + frameTime / 2) / frameTime);
    if(currentFrame > frame + 1) currentFrame = (int) ((time - startTime - frameTime / 2) / frameTime);

    // update mouse position
    if(Mouse.dx != 0 || Mouse.dy != 0 || Mouse.dwheel != 0){
      mousePosition.x += Mouse.dx;
      mousePosition.y -= Mouse.dy;

      if(mousePosition.x < 0) mousePosition.x = 0;
      if(mousePosition.x >= width) mousePosition.x = width - 1;
      if(mousePosition.y < 0) mousePosition.y = 0;
      if(mousePosition.y >= height) mousePosition.y = height - 1;
    }

    // move objects to current frame
    for(; frame < currentFrame; frame++) moveObjects();
  }

  // hope that we don't need to send a packet
  private DataOutputStream packetOutputStream = null;

  private void getPacketOutputStream() throws IOException{
    if(packetOutputStream != null) return;
    packetOutputStream = network.getPacketOutputStream();

    packetOutputStream.writeLong(frame);
  }

  private void sendGameObject(GameObject gameObject){
    try{
      getPacketOutputStream();

      packetOutputStream.writeInt(gameObject.getClassID());
      packetOutputStream.writeInt(gameObject.objectID);
      packetOutputStream.writeBoolean(gameObject.hasAction());

      gameObject.sendUpdate(packetOutputStream);
    } catch(IOException e){
      e.printStackTrace();
    }

    gameObject.lastUpdateFrame = frame;
  }

  private void moveObjects(){
    // recieved packets
    processPackets();

    // send packets that need update
    boolean oldObjects = false;

    Iterator sendObjectIterator = gameObjects.iterator();
    while(sendObjectIterator.hasNext()){
      GameObject gameObject = ((GameObject) sendObjectIterator.next());
      if(gameObject.isHologram()) continue;

      // send objects that want to do some action
      if(gameObject.hasAction() || gameObject.needsUpdate()){
        gameObject.setNeedsUpdate(false);
        sendGameObject(gameObject);
      }

/*
      // remember if we have very old objects
      if(gameObject.lastUpdateFrame < frame - 120){
        oldObjects = true;
      }
*/
    }

    if(packetOutputStream != null || oldObjects){
      // if we send a packet we can fill it with objects that will need update soon
      Iterator oldObjectsIterator = gameObjects.iterator();
      while(oldObjectsIterator.hasNext() && (packetOutputStream == null || network.getBytesLeft() > 200)){
        GameObject gameObject = ((GameObject) oldObjectsIterator.next());
        if(gameObject.isHologram()) continue;

        if(gameObject.lastUpdateFrame < frame - 30){
          sendGameObject(gameObject);
        }
      }

      network.sendPacket();
      packetOutputStream = null;
    }

    // doAction
    Iterator doActionIterator = gameObjects.iterator();
    while(doActionIterator.hasNext()){
      GameObject gameObject = ((GameObject) doActionIterator.next());

      if(gameObject.hasAction()){
        gameObject.setHasAction(false);
        gameObject.doAction();
      }
    }

    // update the game objects list
    Iterator gameObjectsAddIterator = gameObjectsAdd.iterator();
    while(gameObjectsAddIterator.hasNext()){
      GameObject gameObject = (GameObject) gameObjectsAddIterator.next();
      sendGameObject(gameObject);
    }

    // add
    gameObjects.addAll(gameObjectsAdd);
    objectFinder.addAll(gameObjectsAdd);
    gameObjectsAdd.clear();

    // remove and dispose
    objectFinder.removeAll(gameObjectsRemove);

    Iterator removeObjectsIterator = gameObjectsRemove.iterator();
    while(removeObjectsIterator.hasNext()){
      ((GameObject) removeObjectsIterator.next()).dispose();
    }

    gameObjects.removeAll(gameObjectsRemove);
    gameObjectsRemove.clear();

    // move objects
    Iterator doMoveIterator = gameObjects.iterator();
    while(doMoveIterator.hasNext()){
      GameObject gameObject = ((GameObject) doMoveIterator.next());
      gameObject.updateLastPosition();
      gameObject.doMove();

      objectFinder.moved(gameObject);
    }
  }

  /**
   * Adds an object to the game.
   *
   * @param object GameObject
   */
  public void addGameObject(GameObject object){
    gameObjectsAdd.add(object);

    object.objectID = nextObjectID;
    nextObjectID++;
  }

  /**
   * Removes an object from the game.
   *
   * @param object GameObject
   */
  public void removeGameObject(GameObject object){
    gameObjectsRemove.add(object);
  }

  /**
   * Returns all game objects within the given distance from position.
   *
   * @param position position
   * @param distance distance
   * @return list of game objects
   */
  public ArrayList getGameObjects(Point2D.Float position, float distance){
    return objectFinder.getGameObjects(position, distance);
  }

  /**
   * Scrolls the given point to visible
   *
   * @param x x
   * @param y y
   */
  public void scrollTo(float x, float y){
    float scrollToX = -x + Display.getWidth() / 2;
    float scrollToY = -y + Display.getHeight() / 2;

    if(scrollToX > 0) scrollToX = 0;
    if(scrollToY > 0) scrollToY = 0;
    if(-scrollToX + Display.getWidth() > width) scrollToX = Display.getWidth() - width;
    if(-scrollToY + Display.getHeight() > height) scrollToY = Display.getHeight() - height;

    scrollToPoint = new Point2D.Float(scrollToX, scrollToY);
  }

  /**
   * Does the actual scrolling of the game field.
   */
  public void doScroll(){
    // update game field scroll position
    float dx = scrollToPoint.x - position.x;
    float dy = scrollToPoint.y - position.y;
    float distance = (float) (Math.sqrt(Math.pow(dx, 2) + Math.pow(dy, 2)));
    if(distance > 0.01){
      scrollSpeed = (scrollSpeed + distance) / 10;

      position.x += dx * scrollSpeed / distance;
      position.y += dy * scrollSpeed / distance;
    }

    position.x += 25 * (Math.random() - 0.5) * shakeStrength;
    position.y += 25 * (Math.random() - 0.5) * shakeStrength;
    shakeStrength *= 0.98;
  }

  /**
   * Lets the game field shake.
   *
   * @param shakeStrength 0..1
   */
  public void shake(float shakeStrength){
    this.shakeStrength = Math.min(1, this.shakeStrength + shakeStrength);
  }

  /**
   * Returns the width of the game field.
   *
   * @return width
   */
  public int getWidth(){
    return width;
  }

  /**
   * Returns the height of the game field.
   *
   * @return height
   */
  public int getHeight(){
    return height;
  }

  /**
   * Returns the screen position for the given field position.
   *
   * @param x x (field coordinate)
   * @param y y (field coordinate)
   * @return screen position
   */
  public Point fieldToScreen(int x, int y){
    return new Point(x + (int) position.x, y + (int) position.y);
  }

  /**
   * Returns the field position for the given screen position.
   *
   * @param x x (screen coordinate)
   * @param y y (screen coordinate)
   * @return field position
   */
  public Point screenToField(int x, int y){
    return new Point(x - (int) position.x, y - (int) position.y);
  }

  /**
   * Returns the dot map value at the given position.
   *
   * @param x x
   * @param y y
   * @return dot map value
   */
  public byte getDot(int x, int y){
    if(x < 0 || x >= width || y < 0 || y >= height) return DOTMAP_BACKGROUND;
    return dotMap[x + y * width];
  }

  /**
   * Sets the dot at the given position.
   *
   * @param x x
   * @param y y
   * @param dot dot
   */
  public void setDot(int x, int y, byte dot){
    if(x < 0 || x >= width || y < 0 || y >= height) return;
    dotMap[x + y * width] = dot;
  }

  /**
   * Returns the degree of hardness of the point (0..1).
   *
   * @param x x
   * @param y y
   * @return  0..1 (background .. very hard)
   */
  public float getHardness(int x, int y){
    if(x < 0 || x >= width || y < 0 || y >= height) return 0;

    byte value = dotMap[x + y * width];

    if(value == DOTMAP_BACKGROUND) return 0;
    if(DOTMAP_GROUND_WEAK <= value && value <= DOTMAP_GROUND_HARD){
      return (value - DOTMAP_GROUND_WEAK + 1) / (float) (DOTMAP_GROUND_HARD - DOTMAP_GROUND_WEAK + 1);
    }

    return 0;
  }

  /**
   * Returns the foreground texture.
   *
   * @return foreground
   */
  public BigTexture getForegroundTexture(){
    return landscapeFront;
  }

  /**
   * Returns the background texture.
   *
   * @return background
   */
  public BigTexture getBackgroundTexture(){
    return landscapeBack;
  }

  /**
   * Returns the position of the mouse.
   *
   * @return mouse position
   */
  public Point getMousePosition(){
    return mousePosition;
  }

  /**
   * Sets the mouse position.
   *
   * @param x x
   * @param y y
   */
  public void setMousePosition(int x, int y){
    mousePosition.x = x;
    mousePosition.y = y;
  }

  /**
   * Draws the background of the game field.
   */
  public void drawBack(){
    landscapeBack.draw((int) position.x, (int) position.y);
  }

  /**
   * Draws the foreground of the game field.
   */
  public void drawFront(){
    landscapeFront.draw((int) position.x, (int) position.y);
  }

  /**
   * Draws the objects within the game.
   */
  public void drawObjects(){
    double timeDifference = getTime() - getTime(frame + 1);
    // position = new + factor*(new - old)
    float oldFactor = (float) (timeDifference / frameTime);
    float newFactor = 1 + oldFactor;

    // clone game object for thread safety
    ArrayList gameObjectsClone = (ArrayList) gameObjects.clone();

    // move objects
    Iterator gameObjectsIterator = gameObjectsClone.iterator();
    while(gameObjectsIterator.hasNext()){
      GameObject gameObject = (GameObject) gameObjectsIterator.next();
      gameObject.setInterpolatedPosition(newFactor * gameObject.getPosition().x - oldFactor * gameObject.getLastPosition().x,
                                         newFactor * gameObject.getPosition().y - oldFactor * gameObject.getLastPosition().y);
      gameObject.draw();
    }
  }

// network communication
  private synchronized void processPackets(){
    while(!recievedPackets.isEmpty()){
      DataInputStream in = (DataInputStream) recievedPackets.dequeue();

      try{
        long updateFrame = in.readLong();

        // read all objects
        while(in.available() > 0){
          int classID = in.readInt();
          if(classID == 0) break;
          int objectID = in.readInt();
          // if the update is action
          boolean action = in.readBoolean();

          System.out.println("ClassID: " + classID);
          System.out.println("ObjectID: " + objectID);

          GameObject hologram = (GameObject) holograms.get(new Integer(objectID));

          // create or update
          if(hologram == null){
            hologram = ((GameObject) id2Class.get(new Integer(classID))).create(this, in);
            hologram.objectID = objectID;

            holograms.put(new Integer(objectID), hologram);
            gameObjects.add(hologram);
          } else{
            hologram.recieveUpdate(in);
          }

          hologram.setFrame(updateFrame);
          if(action) hologram.setHasAction(true);
        }
      } catch(IOException e){
        System.out.println("Error reading from update packet.");
      }
    }
  }

  public void gameJoined(int id){
  }

  public synchronized void receivePacket(DataInputStream in){
    recievedPackets.enqueue(in);
  }

  public void foundGame(GameInfo gameInfo){
  }
}
