
package jworms.game;

import java.awt.geom.Point2D;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

public abstract class GameObject{
  /** True if the object has the key focus */
  private boolean keyFocus = false;

  /** True if the object wants to do some action. */
  protected boolean hasAction = false;

  /** Last position. */
  private Point2D.Float lastPosition = new Point2D.Float();
  /** Interpolated position. */
  private Point2D.Float interpolatedPosition = new Point2D.Float();

  /** Burn factor. */
  protected float burnFactor = 0;

// network
  /** ID for the object class. */
  public abstract int getClassID();
  public abstract void setClassID(int id);
  /** ID for the object. */
  public int objectID = 0;

  /** True if this object is not a real object, but a hologram project over network. */
  protected boolean isHologram = false;
  /** Current frame of the object (object life time). doMove() increases this value by one. */
  protected long frame = 0;
  /** Last frame the properies were sent to all holograms or the last frame the hologram was updated. */
  protected long lastUpdateFrame = 0;
  /** True if the object wants to send its properties over the network. */
  protected boolean needsUpdate = false;

  /**
   * Constructor for GameObject.
   *
   * @param isHologram true if this is a hologram
   */
  public GameObject(boolean isHologram){
    this.isHologram = isHologram;
  }

  /**
   * Returns the position in float coordinates.
   *
   * @return object position
   */
  public abstract Point2D.Float getPosition();

  /**
   * Stores the current position in lastPositoin. (called by GameField before doMove).
   */
  public void updateLastPosition(){
    lastPosition.x = getPosition().x;
    lastPosition.y = getPosition().y;
  }

  /**
   * Returns the last position of the object (in the last frame).
   *
   * @return last position
   */
  public Point2D.Float getLastPosition(){
    return lastPosition;
  }

  /**
   * Returns the interpolated position of the object.
   *
   * @return interpolated position
   */
  public Point2D.Float getInterpolatedPosition(){
    return interpolatedPosition;
  }

  /**
   * Sets the interpolated position (called by GameField.drawObjects()).
   *
   * @param x x
   * @param y y
   */
  public void setInterpolatedPosition(float x, float y){
    interpolatedPosition.x = x;
    interpolatedPosition.y = y;
  }

  /**
   * Returns true if the object wants to do some action.
   *
   * @return true if action is pending
   */
  public boolean hasAction(){
    return hasAction;
  }

  /**
   * Sets if this object wants to do some action.
   *
   * @param hasAction true if action is pending
   */
  public void setHasAction(boolean hasAction){
    this.hasAction = hasAction;
  }

  /**
   * Fires the action. (this method is called after doMove(), if hasAction() is true)
   */
  public abstract void doAction();

  /**
   * Moves the object without doing actions (changing the environment).
   */
  public abstract void doMove();

  /**
   * Draws the game object.
   */
  public abstract void draw();

  /**
   * Free all resources associated with this object.
   */
  public abstract void dispose();

  /**
   * Returns true if the object has the key focus.
   *
   * @return true if the object has the key focus
   */
  public boolean hasKeyFocus(){
    return keyFocus;
  }

  /**
   * Sets if the object has the key focus.
   *
   * @param keyFocus true if the object has key focus
   */
  public void setKeyFocus(boolean keyFocus){
    this.keyFocus = keyFocus;
  }

// object interaction
  /**
   * Called if the object is hit. (hit strength = motion * weight)
   * (The weight of a normal player is 1.)
   *
   * @param motion motion vector of the hit
   * @param weight weight of the object that does the hit
   */
  public void hit(Point2D.Float motion, float weight){
  }

  /**
   * Called if the object is burned.
   *
   * @param intensity intensity 0..1
   */
  public void burn(float intensity){
    burnFactor = Math.min(1, burnFactor + intensity);
  }

// network
  /**
   * Returns true if this object is a hologram.
   *
   * @return true if this object is a hologram
   */
  public boolean isHologram(){
    return isHologram;
  }

  /**
   * Returns the current frame of the object (object life time).
   *
   * @return current frame
   */
  public long getFrame(){
    return frame;
  }

  /**
   * Sets the frame of the object.
   *
   * @param frame current frame
   */
  public void setFrame(long frame){
    this.frame = frame;
  }

  /**
   * Retruns true if the object wants to send its properties over the network.
   *
   * @return true if this object needs to send an update
   */
  public boolean needsUpdate(){
    return needsUpdate;
  }

  /**
   * Sets if this object needs an update.
   *
   * @param needsUpdate true if needs update
   */
  public void setNeedsUpdate(boolean needsUpdate){
    this.needsUpdate = needsUpdate;
  }

  /**
   * Creates a object and reads all properties from the given input stream.
   * The created object will be a hologram.
   *
   * @param gameField game field
   * @param in input stream
   * @return GameObject
   */
  public abstract GameObject create(GameField gameField, DataInputStream in) throws IOException;

  /**
   * If this is NOT A HOLOGRAM, sends an update (writes all properties to the stream).
   *
   * @param out output stream to write to
   */
  public abstract void sendUpdate(DataOutputStream out) throws IOException;

  /**
   * If this is A HOLOGRAM, recieves an update of all properties.
   *
   * @param in input stream to read from
   */
  public abstract void recieveUpdate(DataInputStream in) throws IOException;
}
