
package jworms.game;

import jworms.game.weapons.Weapon;
import org.lwjgl.input.Keyboard;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

/**
 * Abstract super class of all players.
 */
abstract public class Player extends GameMovingObject{
// static
  /** State: movement direction LEFT. */
  public static final byte DIRECTION_LEFT = 0;
  /** State: movement direction RIGHT. */
  public static final byte DIRECTION_RIGHT = 1;

  /** Key order for holding key state within one int. */
  public static final int KEY_STATE_CONFIG[] = {
    GameConfiguration.ACTION_MOVE_LEFT,
    GameConfiguration.ACTION_MOVE_RIGHT,
    GameConfiguration.ACTION_JUMP,
    GameConfiguration.ACTION_FIRE,
    GameConfiguration.ACTION_AIM_UP,
    GameConfiguration.ACTION_AIM_DOWN,
  };
  /** Key masks according to KEY_STATE_CONFIG. */
  public static final int KEY_MASK_MOVE_LEFT    = 1 << 0;
  public static final int KEY_MASK_MOVE_RIGHT   = 1 << 1;
  public static final int KEY_MASK_JUMP         = 1 << 2;
  public static final int KEY_MASK_FIRE         = 1 << 3;
  public static final int KEY_MASK_AIM_UP       = 1 << 4;
  public static final int KEY_MASK_AIM_DOWN     = 1 << 5;

  /**
   * Loads textures and sound.
   */
  public static void init(){
    PlayerUI.init();
  }

// non static
  /** Name of the player. */
  private String name = null;
  /** Game field. */
  protected GameField gameField = null;

  /** Current key state according to KEY_STATE_CONFIG. */
  private int keyState = 0;

  /** Health of the worm */
  protected float health = 1;
  /** Energy of the worm */
  protected float energy = 1;

  /** Current weapon of the player. */
  private Weapon weapon = null;

  /** Angle if the worm is aiming. */
  private float aimingAngle = 0;

  /** True if the player is charging weapon strength. */
  private boolean isCharging = false;
  /** Charging value if we want to shoot (shoot strength). */
  private float chargingValue = 0;

  /**
   * Creates a player.
   *
   * @param isHologram true if this is a hologram
   * @param name name of the player
   * @param gameField game field
   */
  public Player(boolean isHologram, String name, GameField gameField){
    super(isHologram);

    this.name = name;
    this.gameField = gameField;
  }

  /**
   * Returns the name of the player.
   *
   * @return name
   */
  public String getName(){
    return name;
  }

  /**
   * Should return the health value of the player (normal range [0,1]).
   *
   * @return health value
   */
  public float getHealth(){
    return health;
  }

  /**
   * Returns the energy value of the player (normal range [0,1]).
   *
   * @return energy value
   */
  public float getEnergy(){
    return energy;
  }

  /**
   * Returns the looking direction of the player.
   *
   * @return looking direction
   */
  public abstract int getLookDirection();

  /**
   * Returns the current weapon.
   *
   * @return weapon
   */
  public Weapon getWeapon(){
    return weapon;
  }

  /**
   * Sets the current weapon.
   *
   * @param weapon weapon
   */
  public void setWeapon(Weapon weapon){
    this.weapon = weapon;
  }

  /**
   * Returns true if the player holds a weapon with aim function.
   *
   * @return true if player is aiming
   */
  public boolean isAiming(){
    return weapon != null && weapon.needAiming();
  }

  /**
   * If the player is aiming this method should return the angle to the horizontal.
   *
   * @return aiming angle (radian)
   */
  public float getAimingAngle(){
    return aimingAngle;
  }

  /**
   * Returns true if the player is charging the weapon strength before shoot.
   *
   * @return true if player is charging
   */
  public boolean isCharging(){
    return isCharging;
  }

  /**
   * Returns the charging value if the player is charging the weapon strength.
   *
   * @return charging value
   */
  public float getChargingValue(){
    return chargingValue;
  }

  /**
   * Updates the key state according to KEY_STATE_CONFIG.
   */
  protected void updateKeyState(){
    keyState = 0;
    if(!hasKeyFocus()) return;

    for(int key_nr = 0; key_nr < KEY_STATE_CONFIG.length; key_nr++){
      if(!Keyboard.isKeyDown(GameConfiguration.getKey(KEY_STATE_CONFIG[key_nr]))) continue;
      keyState |= 1 << key_nr;
    }
  }

  /**
   * Returns true if the given key is down.
   *
   * @param keyMask must be one of the KEY_MASK_... constants.
   * @return true if the key is down
   */
  protected boolean isKeyDown(int keyMask){
    return (keyState & keyMask) != 0;
  }

  protected void move(){
    if(!isHologram){
      int lastKeyState = keyState;
      updateKeyState();
      if(keyState != lastKeyState) needsUpdate = true;
    }

    // aim up
    if(isKeyDown(KEY_MASK_AIM_UP)){
      aimingAngle += 0.03;
      if(aimingAngle > Math.PI / 2) aimingAngle = (float) Math.PI / 2;
    }
    // aim down
    if(isKeyDown(KEY_MASK_AIM_DOWN)){
      aimingAngle -= 0.03;
      if(aimingAngle < -Math.PI / 2) aimingAngle = (float) -Math.PI / 2;
    }

    // charge weapon strength / fire weapon
    if(weapon != null) weapon.loop();

    if(weapon != null && weapon.isReady()){
      if(isKeyDown(KEY_MASK_FIRE)){
        if(weapon.needsCharging()){
          isCharging = true;
          chargingValue = Math.min(chargingValue + 0.03f, 1);
        } else{
          fireWeapon();
        }
      } else{
        if(isCharging) fireWeapon();

        isCharging = false;
        chargingValue = 0;
      }
    }
  }

  /**
   * Fires a shoot with the current weapon.
   */
  private void fireWeapon(){
    GameObject firedWeapon = weapon.fireWeapon(this, gameField);
    gameField.addGameObject(firedWeapon);
  }

// network
  public void sendUpdate(DataOutputStream out) throws IOException{
    super.sendUpdate(out);

    out.writeUTF(name);
    out.writeInt(keyState);
    out.writeFloat(health);
    out.writeFloat(energy);
    out.writeFloat(aimingAngle);
  }

  public void recieveUpdate(DataInputStream in) throws IOException{
    super.recieveUpdate(in);

    name = in.readUTF();
    keyState = in.readInt();
    health = in.readFloat();
    energy = in.readFloat();
    aimingAngle = in.readFloat();
  }
}
