
package jworms.game.explosion;

import jworms.game.GameField;
import jworms.game.GameFieldUtil;
import jworms.game.GameObject;
import jworms.openAL.OpenALSource;
import jworms.openAL.OpenALWave;
import jworms.util.MathUtil;
import org.lwjgl.opengl.GL;
import org.lwjgl.openal.AL;

import java.awt.geom.Point2D;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;

public class Explosion extends GameObject{
// static
  /** ID for the object class. */
  private static int classID = 0;
  public int getClassID() { return classID; }
  public void setClassID(int id) { classID = id; }

  /** Explode sound. */
  protected static OpenALWave soundExplode = null;

  /**
   * Loads textures and sound.
   */
  public static void init(){
    ExplosionCloud.init();
    ExplosionFireball.init();
    ExplosionDirtyFireball.init();
    ExplosionFirering.init();

    // load sound
    try{
      soundExplode = OpenALWave.create("sounds/Explosion.wav");
    } catch(Exception e){
      System.out.println("Error in Explosion during loading sound.");
      e.printStackTrace();
    }

    // register class id
    GameField.registerClass(new Explosion());
  }

// non static
  /** GameField */
  private GameField gameField = null;
  /** Position of the explosion. */
  private Point2D.Float position = null;

  /** Time when the exlosion started. */
  private long startTime = 0;
  /** Seconds of livetime. */
  protected float lifeSeconds = 0;

  /** ExplosionObjects */
  private ArrayList explosionObjects = new ArrayList();
  /** Count of dirty fireballs */
  private int dirtyFireballsCount = 0;

  /** Sound source. */
  private OpenALSource soundSource = new OpenALSource();

  /**
   * Creates an explosion.
   *
   * @param isHologram true if this is a hologram
   * @param gameField game field
   * @param position position
   */
  public Explosion(boolean isHologram, GameField gameField, Point2D.Float position){
    super(isHologram);

    this.gameField = gameField;
    this.position = position;

    startTime = System.currentTimeMillis();

    // initialize explosion
    explosionObjects.add(new ExplosionFireball(gameField, position, new Point2D.Float()));

    dirtyFireballsCount = 0;

    // play sound
    if(soundSource.bind()){
      soundSource.stop();
      soundSource.setPos(position.x, position.y, 0);
      soundSource.setPropf(AL.AL_GAIN, 1);
      soundSource.play(soundExplode);
    }

    // modify game field
    if(!isHologram){
      GameFieldUtil.doExplosion(gameField, (int) position.x, (int) position.y, 50, 5);
    }
  }

  /**
   * Creates a explosion that will not be usable as game object.
   */
  public Explosion(){
    super(true);
  }

  public Point2D.Float getPosition(){
    return position;
  }

  public void doAction(){
    gameField.removeGameObject(this);
  }

  public void doMove(){
    lifeSeconds = (System.currentTimeMillis() - startTime) / 1000.f;

    Iterator explosionObjectsIterator = explosionObjects.iterator();
    while(explosionObjectsIterator.hasNext()){
      ExplosionObject explosionObject = (ExplosionObject) explosionObjectsIterator.next();
      explosionObject.doMove();

      if(explosionObject.isFinished()) explosionObjectsIterator.remove();
    }

    if(dirtyFireballsCount < lifeSeconds * 10 && lifeSeconds < 0.5){
      float size = (float) (50 * Math.pow(lifeSeconds, 0.3));

      Point2D.Float objectMotion = new Point2D.Float(
              MathUtil.inBounds(-0.25f, (float) Math.random() - 0.5f, 0.25f),
              MathUtil.inBounds(-0.25f, (float) Math.random() - 0.5f, 0.25f));
      Point2D.Float objectPosition = new Point2D.Float(position.x + objectMotion.x * size, position.y + objectMotion.y * size);

      ExplosionDirtyFireball dirtyFireball = new ExplosionDirtyFireball(gameField, objectPosition, objectMotion);
      explosionObjects.add(dirtyFireball);

      dirtyFireballsCount++;
    }

    // dispose explosion
    if(!isHologram && lifeSeconds > 3) hasAction = true;
  }

  public void draw(){
    GL.glDisable(GL.GL_ALPHA_TEST);
    GL.glEnable(GL.GL_BLEND);

    Iterator explosionObjectsIterator = explosionObjects.iterator();
    while(explosionObjectsIterator.hasNext()){
      ExplosionObject explosionObject = (ExplosionObject) explosionObjectsIterator.next();
      explosionObject.draw();
    }

    GL.glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
  }

  public void dispose(){
    soundSource.dispose();
  }

// network
  public GameObject create(GameField gameField, DataInputStream in) throws IOException{
    Explosion explosion = new Explosion(true, gameField, new Point2D.Float());
    explosion.recieveUpdate(in);
    return explosion;
  }

  public void sendUpdate(DataOutputStream out) throws IOException{
    out.writeFloat(position.x);
    out.writeFloat(position.y);
  }

  public void recieveUpdate(DataInputStream in) throws IOException{
    position.x = in.readFloat();
    position.y = in.readFloat();
  }
}
