
package jworms.game.weapons;

import jworms.game.GameField;
import jworms.game.GameObject;
import jworms.game.Player;
import jworms.openGL.SmallTexture;
import org.lwjgl.opengl.GL;

import java.awt.*;
import java.awt.geom.Point2D;

public class DesertEagle implements Weapon{
// static
  /** Texture for the weapon. */
  private static SmallTexture weaponTexture = null;

  /**
   * Loads textures and sound.
   */
  public static void init(){
    // load images
    try{
      weaponTexture = new SmallTexture("images/weapons/deserteagle/deserteagle.png", GL.GL_LINEAR);
    } catch(Exception e){
      System.out.println("Error in DesertEagle during loading images.");
      e.printStackTrace();
    }
  }

// non static
  /** Game field. */
  protected GameField gameField = null;

  /** Center of the weapon in texture coordinates. */
  protected Point relativeWeaponCenter = new Point(11, 16);
  /** Position of the weapon center relative to the position of the Player. */
  protected Point relativeWeaponCenterPosition = new Point(14, 22);
  /** Distance between the weapon center and the barrel center. */
  protected int gripBarrelDistance = -3;

  /** Loops till next shoot can be fired. */
  protected int loopsTillReady = 0;
  /** Reload process: 1 means ready. */
  protected float reloadProcess = 0;

  /** Damage radius of shoots. */
  protected int damageRadius = 10;

  public DesertEagle(GameField gameField){
    this.gameField = gameField;
  }

  /**
   * Returns the texture for the weapon.
   *
   * @return weapon texture
   */
  public SmallTexture getTexture(){
    return weaponTexture;
  }

  /**
   * Returns a shoot angle with a little bit random noise.
   *
   * @param player player
   * @return shoot angle
   */
  protected float getShootAngle(Player player){
    return player.getAimingAngle() + ((float) Math.random() - 0.5f) / 30;
  }

  public GameObject fireWeapon(Player player, GameField gameField){
    if(!isReady()) return null;

    // loops till next shoot
    loopsTillReady = 30;

    // calculate bullet
    float speed = 30;

    float angle = getShootAngle(player);
    float dx = (float) Math.cos(angle) * (player.getLookDirection() == Player.DIRECTION_RIGHT ? 1 : -1);
    float dy = (float) -Math.sin(angle);

    Point position = getCenter(player);
    position.x += dx * 10;
    position.y += dy * 10;

    Point2D.Float motion = new Point2D.Float(dx * speed, dy * speed);

    return new Bullet(false, gameField, new Point2D.Float(position.x, position.y), motion, damageRadius);
  }

  public void drawSymbol(Player player){
    int directionFactor = player.getLookDirection() == Player.DIRECTION_RIGHT ? 1 : -1;

    Point position = getGripCenter(player);
    position.x -= relativeWeaponCenter.x;
    position.y -= relativeWeaponCenter.y;

    float angle = player.getAimingAngle() * directionFactor;
    Point screen = gameField.fieldToScreen(position.x, position.y);
    getTexture().draw(screen.x, screen.y, relativeWeaponCenter.x, relativeWeaponCenter.y, player.getLookDirection() == Player.DIRECTION_RIGHT ? 1 : -1, 1, (float) (angle * 180 / Math.PI));
  }

  /**
   * Returns the aiming offset (barrel center).
   *
   * @param player player
   * @return barrel center
   */
  public Point getCenter(Player player){
    return getGripCenter(player, gripBarrelDistance);
  }

  /**
   * Returns the rotation center of the texture (grip point).
   *
   * @param player player
   * @return grib point
   */
  private Point getGripCenter(Player player){
    return getGripCenter(player, 0);
  }

  /**
   * Returns a point vertically moved from grip point.
   *
   * @param player player
   * @param gribDistance distance from grip point
   * @return moved grip point
   */
  private Point getGripCenter(Player player, int gribDistance){
    int directionFactor = player.getLookDirection() == Player.DIRECTION_RIGHT ? 1 : -1;
    float angle = player.getAimingAngle() * directionFactor;

    int x = (int) player.getInterpolatedPosition().x + (relativeWeaponCenterPosition.x + (int) (Math.sin(angle) * gribDistance)) * directionFactor;
    int y = (int) player.getInterpolatedPosition().y + relativeWeaponCenterPosition.y + (int) (Math.cos(angle) * gribDistance);

    return new Point(x - relativeWeaponCenter.x * directionFactor, y - relativeWeaponCenter.y);
  }

  public boolean needAiming(){
    return true;
  }

  public boolean needsCharging(){
    return false;
  }

  public void loop(){
    reloadProcess = Math.min(reloadProcess + 0.02f, 1);
    if(loopsTillReady > 0) loopsTillReady--;
  }

  public boolean isReady(){
    return loopsTillReady == 0 && reloadProcess == 1;
  }

  public float getReloadProcess(){
    return reloadProcess;
  }
}
