
package jworms.network;

import java.io.*;
import java.net.*;
import java.util.ArrayList;
import java.util.HashMap;

public class UDPNetwork {

// static
  /** Port, on which hosts will listen to. */
  public static final int HOST_PORT = 8000;
  /** Port, on which games will be played. */
  public static final int FIRST_GAME_PORT = HOST_PORT + 1;
  /** Last stored packets count. */
  private static final int STORED_PACKETS_COUNT = 50;
  /** Packet signature of all packet send by this application. */
  public static final String SIGNATURE = "jwo";
  public static final int TYPE_GAME_INFO = 0;
  public static final int TYPE_GET_GAMES = 1;
  public static final int TYPE_JOIN = 2;
  public static final int TYPE_DATA = 3;
  public static final int TYPE_YOUR_ID = 4;
  public static final int TYPE_LOST_PACKETS = 5;

  /**
   * Writes the "jwo"-SIGNATURE to the output stream.
   *
   * @param out the output stream
   * @throws IOException thown if writing SIGNATURE fails
   */
  public static void writeSignature(DataOutputStream out) throws IOException {
    out.writeBytes(SIGNATURE);
  }

  /**
   * Reads the 3 byte SIGNATURE from the input stream and returns true if it's equal to "jwo"
   *
   * @param in the input stream
   * @return true if the SIGNATURE of this stream is equal to "jwo"
   * @throws IOException thown if reading SIGNATURE fails
   */
  public static boolean readSignature(DataInputStream in) throws IOException {
    byte[] sig = new byte[SIGNATURE.length()];
    in.read(sig, 0, SIGNATURE.length());
    return SIGNATURE.equals(new String(sig));
  }


// non static
  /** Game info of the running game. */
  private GameInfo gameInfo;
  /** My player id. -1 if the player ist not in a game. */
  private int id = -1;
  /** Thread listening on HOST_PORT to tell other that here runs a game. */
  private UDPHost udpHost = null;
  /** Thread listening on game port to receive game packet from other players. */
  private UDPReceiver udpReceiver = null;
  /** Thread sending "get games"-request to all hosts an listening to there reply. */
  private UDPGetGames udpGetGames = null;
  /** Socket used to inform other players in this game. */
  private DatagramSocket socket;
  /** Boradcast address to send the packets to. */
  private InetAddress iaddr;
  /** Data which is send by sendPacket(). */
  private ByteArrayOutputStream out;
  /** Size of an udp packet. */
  private int packetSize;
  /** Contains the IP addresses of players asked for joining and there playerIDs. */
  private HashMap playerIPs;
  /** Number of the last packet for each player. */
  private int lastPacketNo[];
  /** List of lost packets for each player in the game. */
  private ArrayList[] lostPackets;
  /** Count of received packets after lost one for each player. */
  private int[] packetsAfterLost;
  /** Last 50 packet I've sent. */
  private byte[][] myLastPackets;

  /** Listeners: udpNetworkListeners of type UDPNetworkListener. */
  private ArrayList udpNetworkListeners = new ArrayList();

  /**
   * UDPNetworkListener can register themselves at UDPNetwork
   * to be informed about UDPNetwork events.
   *
   * @param listener listener who wants to be informed about changes
   */
  public void addUDPNetworkListener(UDPNetworkListener listener) {
    if(udpNetworkListeners.contains(listener)) return;
    udpNetworkListeners.add(listener);
  }

  /**
   * UDPNetworkListener can sign off at UDPNetwork
   * to be no longer informed about UDPNetwork events.
   *
   * @param listener listener who wants to sign off
   */
  public void removeUDPNetworkListener(UDPNetworkListener listener) {
    udpNetworkListeners.remove(listener);
  }

  /**
   * Opens a game.
   *
   * @param gameName name of the game.
   * @param maxPlayerCount maximal count of players
   * @throws SocketException thrown if opening the game failed
   */
  public void openGame(String gameName, int maxPlayerCount) throws SocketException {
    gameInfo = new GameInfo(gameName, 1, maxPlayerCount, FIRST_GAME_PORT);
    // interrupt "looking for games" and create socket
    initNetwork();

    // add me as player 0
    try{
      id = 0;
      playerIPs = new HashMap(maxPlayerCount);
      playerIPs.put(InetAddress.getLocalHost(), new Integer(id));
    } catch(UnknownHostException e){
    }

    udpHost = new UDPHost(gameInfo, HOST_PORT);
  }

  /**
   * Joins a game.
   *
   * @param gameInfo the game to join
   * @throws IOException thrown if joining the game failed
   */
  public void joinGame(GameInfo gameInfo) throws IOException {
    this.gameInfo = gameInfo;
    initNetwork();

    // send join command
    ByteArrayOutputStream byteOut = new ByteArrayOutputStream();
    getPacketOutputStream(byteOut, TYPE_JOIN);
    byte[] bytes = byteOut.toByteArray();
    DatagramPacket p = new DatagramPacket(bytes, bytes.length, iaddr, gameInfo.getPort());
    sendPacket(p);
  }

  /**
   * Call this methode to prepare for the game. It interrupts the "looking for games" thread,
   * creates the socket, initialized variables and starts the udpReceiver.
   *
   * @throws SocketException thrown if creating the socket failed
   */
  private void initNetwork() throws SocketException {
    // abort looking for games
    if(udpGetGames != null){
      udpGetGames.interrupt();
      udpGetGames = null;
    }

    lastPacketNo = new int[gameInfo.getMaxPlayerCount()];
    lostPackets = new ArrayList[gameInfo.getMaxPlayerCount()];
    for(int i = 0; i < lostPackets.length; i++){
      lostPackets[i] = new ArrayList();
    }
    packetsAfterLost = new int[gameInfo.getMaxPlayerCount()];
    myLastPackets = new byte[STORED_PACKETS_COUNT][];

    createSocket();

    udpReceiver = new UDPReceiver(this, gameInfo.getPort());
  }

  /**
   * Creates te socket used to send packets to other players.
   *
   * @throws SocketException thrown if creating the socket failed.
   */
  private void createSocket() throws SocketException {
    socket = new DatagramSocket();
    packetSize = socket.getSendBufferSize();
    try{
      iaddr = InetAddress.getByName("255.255.255.255");
    } catch(UnknownHostException e){
      System.out.println("Error during creating socket: " + e.getMessage());
    }
  }

  /**
   * Leaves the current game.
   */
  public void leaveGame() {
    id = -1;
    // stop listening on HOST_PORT
    if(udpHost != null){
      udpHost.interrupt();
      udpHost = null;
    }
    if(udpReceiver != null){
      udpReceiver.interrupt();
      udpReceiver = null;
    }
    // stop sending packets
    socket.close();
  }

  /**
   * Looks for running games and informs the listeners wenn found same.
   */
  public void lookForGames() {
    try{
      if(udpGetGames == null){
        // listening reply
        udpGetGames = new UDPGetGames(this, HOST_PORT);
        // wait for thread
        try{
          Thread.sleep(200);
        } catch(InterruptedException e){
        }
      }
      // sending request
      udpGetGames.sendRequest();
    } catch(IOException e){
      System.out.println("Error looking for games: " + e.getMessage());
    }
  }

  public DataOutputStream getPacketOutputStream() throws IOException {
    out = new ByteArrayOutputStream(socket.getSendBufferSize());
    // write header
    DataOutputStream dataOut = getPacketOutputStream(out, TYPE_DATA);  // type
    dataOut.writeByte(id);  // player id
    dataOut.writeInt(++lastPacketNo[id]);
    return dataOut;
  }

  protected DataOutputStream getPacketOutputStream(ByteArrayOutputStream byteOut, int type) throws IOException {
    DataOutputStream dataOut = new DataOutputStream(byteOut);

    // write header
    writeSignature(dataOut);
    dataOut.writeByte(type);

    return dataOut;
  }

  /**
   * Returns the count of left bytes of the packet.
   *
   * @return count of left bytes of the packet
   */
  public int getBytesLeft() {
    return 1400 - out.size();
  }

  /**
   * Sends a game packet.
   */
  public void sendPacket() {
    if(id == -1) return; // because not in a game

    try{
//      System.out.println("Sending packet...");

      InetAddress iaddr = InetAddress.getByName("255.255.255.255");
      byte[] byteArray = out.toByteArray();
      DatagramPacket p = new DatagramPacket(byteArray, byteArray.length, iaddr, gameInfo.getPort());
      // add packet to list of the last 50 packets
      myLastPackets[lastPacketNo[id]%STORED_PACKETS_COUNT] = byteArray;

      sendPacket(p);
    } catch(IOException e){
      System.out.println("Error sending message: " + e.getMessage());
    }
  }

  /**
   * Sends a packet with the given data.
   */
  protected synchronized void sendPacket(DatagramPacket p) throws IOException {
    socket.send(p);
  }

  /**
   * Called wenn a packet received to inform listeners about the packet if it is a game data packet.
   *
   * @param p udp packet
   * @throws IOException thrown if
   */
  protected void receivePacket(DatagramPacket p) throws IOException {
    ByteArrayInputStream byteIn = new ByteArrayInputStream(p.getData());
    DataInputStream in = new DataInputStream(byteIn);

    if(UDPNetwork.readSignature(in)){
      int type = in.readByte();
      if(type == TYPE_DATA){
        int playerID = in.readByte();
        if(playerID != id && id != -1){
          int packetNo = in.readInt();
          boolean forward = false;
          // test if this packet was an lost packet
          int index = lostPackets[playerID].indexOf(new Integer(packetNo));
          if(index >= 0){ // found it
            System.out.println("Lost packet received: " + packetNo);
            lostPackets[playerID].remove(index);
            if(lostPackets[playerID].size() == 0) packetsAfterLost[playerID] = 0;
            forward = true;
          }else{
            System.out.println("Packet received: " + packetNo);
            // test if this packet was the expected packet; add the lost ones to lostPackets
            if(lastPacketNo[playerID] < packetNo){
              for(int i=lastPacketNo[playerID]+1; i<packetNo; i++)
                lostPackets[playerID].add(new Integer(i));
              lastPacketNo[playerID] = packetNo;
              // one more good packet after lost one? => remember
              if(lostPackets[playerID].size() != 0){
                packetsAfterLost[playerID]++;
                if(packetsAfterLost[playerID] % 10 == 0){
                  // send that these packets are missing
                  ByteArrayOutputStream byteOut = new ByteArrayOutputStream();
                  DataOutputStream out = getPacketOutputStream(byteOut, TYPE_LOST_PACKETS);
                  for(int i = 0; i < lostPackets[playerID].size(); i++){
                    int no = ((Integer) lostPackets[playerID].get(i)).intValue();
                    if(lastPacketNo[playerID] - no > STORED_PACKETS_COUNT){ // packet too old => forget it
                      lostPackets[playerID].remove(i);
                      i--;
                    }else{
                      out.writeInt(no);
                    }
                    System.out.println("Sending LOST_PACKETS request: " + no);
                  }
                  out.writeInt(0); // end
                  byte[] bytes = byteOut.toByteArray();
                  p.setData(byteOut.toByteArray());
                  p.setPort(gameInfo.getPort());
                  sendPacket(p);
                }
              }
              forward = true;
            }
          }
          if(forward){
            // inform listeners about received packet
            for(int i = 0; i < udpNetworkListeners.size(); i++){
              UDPNetworkListener listener = (UDPNetworkListener) udpNetworkListeners.get(i);
              listener.receivePacket(in);
            }
          }
        }
      } else if(type == TYPE_LOST_PACKETS){
        int packet;
        while((packet = in.readInt()) > 0){
          if(lastPacketNo[id] - packet <= STORED_PACKETS_COUNT){
            System.out.println("Answering LOST_PACKET request: " + packet);
            byte[] bytes = myLastPackets[packet%STORED_PACKETS_COUNT];
            DatagramPacket pac = new DatagramPacket(bytes, bytes.length, p.getAddress(), gameInfo.getPort());
            sendPacket(pac);
          }
        }
      } else if(type == TYPE_JOIN && id == 0){
        System.out.println("Other player sent join request!");
        // find his id
        Integer hisID = (Integer) playerIPs.get(p.getAddress());
        if(hisID == null){
          hisID = new Integer(playerIPs.size());
          playerIPs.put(p.getAddress(), hisID);
        }

        // send reply with his id
        ByteArrayOutputStream byteOut = new ByteArrayOutputStream();
        DataOutputStream out = getPacketOutputStream(byteOut, TYPE_YOUR_ID);
        out.writeByte(hisID.intValue());
        p.setData(byteOut.toByteArray());
        p.setPort(gameInfo.getPort());
        System.out.println("His id is " + hisID.intValue());
        sendPacket(p);
      } else if(type == TYPE_YOUR_ID){
        id = in.readByte();
        for(int i = 0; i < udpNetworkListeners.size(); i++){
          UDPNetworkListener listener = (UDPNetworkListener) udpNetworkListeners.get(i);
          listener.gameJoined(id);
        }
      }
    }
  }

  /**
   * Called wenn a game was found to inform listeners about a network game.
   *
   * @param gameInfo game info of the game
   */
  protected void foundGame(GameInfo gameInfo) {
    for(int i = 0; i < udpNetworkListeners.size(); i++){
      UDPNetworkListener listener = (UDPNetworkListener) udpNetworkListeners.get(i);
      listener.foundGame(gameInfo);
    }
  }
}
