
package jworms.openGL;

import org.lwjgl.Display;
import org.lwjgl.opengl.GL;

import java.awt.*;
import java.awt.image.BufferedImage;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;

/**
 * Big texture that may be bigger than 256*256.
 */
public class BigTexture{
// static properties
  /** Maximum number of OpenGLInternalTextures per OpenGLTexture. */
  public static int MAX_CACHE_TILES = 30;

  /** RGB with 1 byte per color (3 bytes per pixel). */
  public static final int FORMAT_RGB = OpenGLTextureInternal.FORMAT_RGB;
  /** RGBA with 1 byte per color (4 bytes per pixel). */
  public static final int FORMAT_RGBA = OpenGLTextureInternal.FORMAT_RGBA;

// texture properties
  /** The texture as image. */
  private BufferedImage texture = null;
  /** Graphics to draw to the image. */
  private Graphics2D graphics = null;

  /** Size of the tiles used for this texture. */
  private int tileSize = 0;
  /** Size of the texture in tiles horizontal. */
  private int tilesHorizontal = 0;
  /** Size of the texture in tiles vertical. */
  private int tilesVertical = 0;
  /** Tiles of the texture (big images must be splitted). */
  private BigTextureTile tiles[][] = null;
  /** Tile cache: last used subtextures (BigTextureTile). */
  private ArrayList tilesCache = null;

  /**
   * Load a texture in OpenGL memory.
   *
   * After all textures were loaded, they must be created with OpenGLInternalTexture.glGenTextures().
   *
   * @param path path to the texture image
   * @param format format of the image (BigTexture.FORMAT_...)
   * @param isFinal true if the texture is final (will not be modified)
   */
  public BigTexture(String path, int format, boolean isFinal) throws Exception{
    Image image = (new javax.swing.ImageIcon(path)).getImage();
    if(image == null || image.getWidth(null) == -1){
      throw new Exception("Error in BigTexture due to texture not found: " + path);
    }

    // BufferedImage format
    int bufferedImageFormat = 0;
    switch(format){
      case OpenGLTextureInternal.FORMAT_RGB:
        bufferedImageFormat = BufferedImage.TYPE_3BYTE_BGR;
        break;
      case OpenGLTextureInternal.FORMAT_RGB5:
        bufferedImageFormat = BufferedImage.TYPE_3BYTE_BGR;
        break;
      case OpenGLTextureInternal.FORMAT_COMPRESSED_RGB:
        bufferedImageFormat = BufferedImage.TYPE_3BYTE_BGR;
        break;
      case OpenGLTextureInternal.FORMAT_RGBA:
        bufferedImageFormat = BufferedImage.TYPE_4BYTE_ABGR;
        break;
      default:
        throw new Exception("Error in BigTexture due tu unkown format: " + format);
    }

    // format of the buffer and OpenGL internal format
    int bufferFormat = 0;
    int internalFormat = 0;
    switch(bufferedImageFormat){
      case BufferedImage.TYPE_3BYTE_BGR:
        bufferFormat = OpenGLTextureInternal.FORMAT_RGB;
        internalFormat = isFinal ? OpenGLTextureInternal.FORMAT_COMPRESSED_RGB : OpenGLTextureInternal.FORMAT_RGB5;
        break;
      case BufferedImage.TYPE_4BYTE_ABGR:
        bufferFormat = OpenGLTextureInternal.FORMAT_RGBA;
        internalFormat = OpenGLTextureInternal.FORMAT_RGB5_A1;
        break;
    }

    // extract the image
    texture = new BufferedImage(image.getWidth(null), image.getHeight(null), bufferedImageFormat);
    graphics = texture.createGraphics();
    graphics.drawImage(image, 0, 0, null);

    // calculate tile size
    for(int size_nr = 0; size_nr < OpenGLTextureInternal.SIZES.length; size_nr++){
      tileSize = OpenGLTextureInternal.SIZES[size_nr];
      if(texture.getWidth() <= tileSize && texture.getHeight() <= tileSize){
        break;
      }
    }

    // number of tiles horizontal/ vertical
    tilesHorizontal = (texture.getWidth() + tileSize - 1) / tileSize;
    tilesVertical = (texture.getHeight() + tileSize - 1) / tileSize;

    // extract all tiles
    tiles = new BigTextureTile[tilesHorizontal][tilesVertical];
    for(int tile_x = 0; tile_x < tilesHorizontal; tile_x++){
      for(int tile_y = 0; tile_y < tilesVertical; tile_y++){
        int x = tileSize * tile_x;
        int y = tileSize * tile_y;
        int width = Math.min(tileSize, texture.getWidth() - x);
        int height = Math.min(tileSize, texture.getHeight() - y);

        BigTextureTile textureTile = new BigTextureTile(tile_x, tile_y, x, y, width, height);

        // internal OpenGL texture
        textureTile.textureInternal = new OpenGLTextureInternal(tileSize, tileSize, bufferFormat, internalFormat, isFinal);
        textureTile.textureInternal.buffer = ByteBuffer.allocateDirect(4 * tileSize * tileSize);

        updateByteBuffer(textureTile, texture);

        // try to compress the texture
        if(isFinal) textureTile.textureInternal.compressTexture();

        tiles[tile_x][tile_y] = textureTile;
      }
    }

    // create tiles cache
    int tilesCount = Math.min(MAX_CACHE_TILES, tilesHorizontal * tilesVertical);
    tilesCache = new ArrayList(tilesCount);

    for(int tile_nr = 0; tile_nr < tilesCount; tile_nr++){
      BigTextureTile textureTile = new BigTextureTile(0, 0, 0, 0, tileSize, tileSize);

      // internal OpenGL texture
      textureTile.textureInternal = new OpenGLTextureInternal(tileSize, tileSize, bufferFormat, internalFormat, isFinal);

      tilesCache.add(textureTile);

      // for later generation of the texture
      OpenGLTextureInternal.addToGenTextures(textureTile.textureInternal);
    }

    // for final textures we don't need a graphics to paint
    if(isFinal){
      graphics.dispose();
      texture = null;
    }
  }

  /**
   * Extracts the given portion of the image into a ByteBuffer.
   * The width and height may lay outside of the texture,
   * but the returned buffer will fit the size and the unused area will be filled with alpha.
   *
   * @param textureTile BigTextureTile
   * @param texture texture image
   */
  private void updateByteBuffer(BigTextureTile textureTile, BufferedImage texture){
    ByteBuffer buffer = textureTile.textureInternal.buffer;

    // width and height of the data to grab
    int x = textureTile.location.x;
    int y = textureTile.location.y;
    int width = textureTile.location.width;
    int height = textureTile.location.height;

    // bytes per pixel
    int bytesPerPixel = texture.getType() == BufferedImage.TYPE_3BYTE_BGR ? 3 : 4;

    // data array (but may have wrong size)
    byte data[] = (byte[]) texture.getRaster().getDataElements(x, y, width, height, null);

    buffer.clear();

    if(width == textureTile.textureInternal.width && height == textureTile.textureInternal.height){
      buffer.put(data);
    } else{
      // resize the data array
      for(int texture_y = 0; texture_y < height; texture_y++){
        buffer.position(texture_y * textureTile.textureInternal.width * bytesPerPixel);
        buffer.put(data, texture_y * width * bytesPerPixel, width * bytesPerPixel);
      }
    }

    buffer.rewind();
  }

  /**
   * Returns the Image of the texture.
   *
   * @return Image
   */
  public BufferedImage getImage(){
    return texture;
  }

  /**
   * Returns the graphics context to draw on the texture.
   *
   * @return graphics context
   */
  public Graphics2D getGraphics(){
    return graphics;
  }

  /**
   * Calculates the tiles that intersect this region.
   *
   * @param x x
   * @param y y
   * @param width width
   * @param height height
   * @return Rectangle with left, top, right, bottom = x, y, x + width, y + width
   */
  private Rectangle getTilesBounds(int x, int y, int width, int height){
    int left = Math.max(0, x / tileSize);
    int top = Math.max(0, y / tileSize);
    int right = Math.min(tilesHorizontal - 1, (x + width) / tileSize);
    int bottom = Math.min(tilesVertical - 1, (y + height) / tileSize);

    return new Rectangle(left, top, right - left + 1, bottom - top + 1);
  }

  /**
   * If changes were made to the texture (getGraphics())
   * you must call this function to update the OpenGL texture.
   *
   * @param x x
   * @param y y
   * @param width width
   * @param height height
   */
  public void update(int x, int y, int width, int height){
    // update the byte buffer of all tiles
    Rectangle tiles_bounds = getTilesBounds(x, y, width, height);
    for(int tile_x = 0; tile_x < tiles_bounds.width; tile_x++){
      for(int tile_y = 0; tile_y < tiles_bounds.height; tile_y++){
        updateByteBuffer(tiles[tile_x + tiles_bounds.x][tile_y + tiles_bounds.y], texture);
      }
    }

    // invalidate all tiles that were affected
    Rectangle bounds = new Rectangle(x, y, width, height);
    Iterator cachedTilesIterator = tilesCache.iterator();
    while(cachedTilesIterator.hasNext()){
      BigTextureTile cacheTexture = (BigTextureTile) cachedTilesIterator.next();
      if(bounds.intersects(cacheTexture.location)) cacheTexture.isValid = false;
    }
  }

  /**
   * Draws the texture.
   *
   * @param x x
   * @param y y
   */
  public void draw(int x, int y){
    // portion of the texture that is shown on the screen
    Rectangle tiles_bounds = getTilesBounds(-x, -y, Display.getWidth(), Display.getHeight());
    if(tiles_bounds.width < 0 || tiles_bounds.height < 0) return;

    // mark tiles that are not in cache
    boolean tiles_needed[][] = new boolean[tiles_bounds.width][tiles_bounds.height];
    for(int tile_x = 0; tile_x < tiles_bounds.width; tile_x++){
      for(int tile_y = 0; tile_y < tiles_bounds.height; tile_y++){
        tiles_needed[tile_x][tile_y] = true;
      }
    }

    // draw tiles from cache
    Iterator tilesIterator = tilesCache.iterator();
    while(tilesIterator.hasNext()){
      BigTextureTile cacheTexture = (BigTextureTile) tilesIterator.next();
      if(!cacheTexture.isValid) continue;

      int tileX = cacheTexture.tileX;
      int tileY = cacheTexture.tileY;

      if(tileX >= tiles_bounds.x && tileX < tiles_bounds.x + tiles_bounds.width &&
              tileY >= tiles_bounds.y && tileY < tiles_bounds.y + tiles_bounds.height){
        tiles_needed[tileX - tiles_bounds.x][tileY - tiles_bounds.y] = false;
        cacheTexture.lastUse = System.currentTimeMillis();

        drawOpenGLTextureInternal(cacheTexture.textureInternal, x + cacheTexture.location.x, y + cacheTexture.location.y);
      }
    }

    // draw the tiles that were not in cache
    Collections.sort(tilesCache);
    int tiles_cache_index = 0;

    for(int tile_x = 0; tile_x < tiles_bounds.width; tile_x++){
      for(int tile_y = 0; tile_y < tiles_bounds.height; tile_y++){
        if(!tiles_needed[tile_x][tile_y]) continue;

        BigTextureTile textureTile = tiles[tile_x + tiles_bounds.x][tile_y + tiles_bounds.y];

        // get the oldest texture from tiles cache
        BigTextureTile cacheTexture = (BigTextureTile) tilesCache.get(tiles_cache_index);
        tiles_cache_index = (tiles_cache_index + 1) % tilesCache.size();

        cacheTexture.tileX = textureTile.tileX;
        cacheTexture.tileY = textureTile.tileY;
        cacheTexture.location.x = textureTile.location.x;
        cacheTexture.location.y = textureTile.location.y;

        // get uncached texture
        cacheTexture.textureInternal.buffer = textureTile.textureInternal.buffer;
        cacheTexture.textureInternal.bufferFormat = textureTile.textureInternal.bufferFormat;
        cacheTexture.textureInternal.internalFormat = textureTile.textureInternal.internalFormat;

        // update the OpenGL texture
        cacheTexture.textureInternal.buffer2OpenGL();

        // cache texture is valid
        cacheTexture.isValid = true;

        // draw the texture
        drawOpenGLTextureInternal(cacheTexture.textureInternal, x + cacheTexture.location.x, y + cacheTexture.location.y);
      }
    }
  }

  /**
   * Draws the OpenGLTextureInternal.
   *
   * @param textureInternal texture
   * @param x x
   * @param y y
   */
  private void drawOpenGLTextureInternal(OpenGLTextureInternal textureInternal, int x, int y){
    int width = textureInternal.width;
    int height = textureInternal.height;

    // draw the texture
    GL.glBindTexture(GL.GL_TEXTURE_2D, textureInternal.nr);
    GL.glBegin(GL.GL_QUADS);
    {
      GL.glTexCoord2f(0.0f, 0.0f);
      GL.glVertex2i(x, y);
      GL.glTexCoord2f(1.0f, 0.0f);
      GL.glVertex2i(x + width, y);
      GL.glTexCoord2f(1.0f, 1.0f);
      GL.glVertex2i(x + width, y + height);
      GL.glTexCoord2f(0.0f, 1.0f);
      GL.glVertex2i(x, y + height);
    }
    GL.glEnd();
  }
}
