
package jworms.openGL;

import org.lwjgl.opengl.GL;

import java.awt.*;
import java.awt.image.BufferedImage;
import java.nio.ByteBuffer;

/**
 * Small texture (<= 256*256) but more flexible than BigTexture.
 */
public class SmallTexture{
// texture properties
  /** The texture as image. */
  private BufferedImage texture = null;
  /** Graphics to draw to the image. */
  private Graphics2D graphics = null;

  /** OpenGL texture. */
  private OpenGLTextureInternal textureInternal = null;

  /** Rectanlge (0, 0, width, height). */
  private Rectangle textureRectangle = null;

  /**
   * Load a texture in OpenGL memory.
   *
   * After all textures were loaded, they must be created with OpenGLInternalTexture.glGenTextures().
   *
   * @param path path to the texture image
   * @param filter interpolation filter (e.a. GL.GL_LINEAR)
   */
  public SmallTexture(String path, int filter) throws Exception{
    Image image = (new javax.swing.ImageIcon(path)).getImage();
    if(image == null || image.getWidth(null) == -1){
      throw new Exception("Texture not found: " + path);
    }

    // extract the image
    texture = new BufferedImage(image.getWidth(null), image.getHeight(null), BufferedImage.TYPE_4BYTE_ABGR);
    graphics = texture.createGraphics();
    graphics.drawImage(image, 0, 0, null);
    textureRectangle = new Rectangle(0, 0, texture.getWidth(), texture.getHeight());

    // calculate the size
    int size = 0;
    for(int size_nr = 0; size_nr < OpenGLTextureInternal.SIZES.length; size_nr++){
      size = OpenGLTextureInternal.SIZES[size_nr];
      if(texture.getWidth() <= size && texture.getHeight() <= size){
        break;
      }
    }

    // internal OpenGL texture
    textureInternal = new OpenGLTextureInternal(size, size, OpenGLTextureInternal.FORMAT_RGBA, OpenGLTextureInternal.FORMAT_RGBA, true);
    textureInternal.buffer = ByteBuffer.allocateDirect(4 * size * size);
    updateByteBuffer();
    textureInternal.filter = filter;

    // for later generation of the texture
    OpenGLTextureInternal.addToGenTextures(textureInternal);
  }

  /**
   * Updates the byte buffer of the texture.
   */
  public void updateByteBuffer(){
    ByteBuffer buffer = textureInternal.buffer;

    int width = texture.getWidth();
    int height = texture.getHeight();

    // data array (but may have wrong size)
    byte data[] = (byte[]) texture.getRaster().getDataElements(0, 0, width, height, null);

    buffer.clear();

    if(width == textureInternal.width && height == textureInternal.height){
      buffer.put(data);
    } else{
      // resize the data array
      for(int texture_y = 0; texture_y < height; texture_y++){
        buffer.position(texture_y * textureInternal.width * 4);
        buffer.put(data, texture_y * width * 4, width * 4);
      }
    }

    buffer.rewind();
  }

  /**
   * Returns the Image of the texture.
   *
   * @return Image
   */
  public BufferedImage getImage(){
    return texture;
  }

  /**
   * Returns the graphics context to draw on the texture.
   *
   * @return graphics context
   */
  public Graphics2D getGraphics(){
    return graphics;
  }

  /**
   * Returns the OpenGL texture.
   *
   * @return OpenGL texture
   */
  public OpenGLTextureInternal getTextureInternal(){
    return textureInternal;
  }

  /**
   * Draws the texture.
   *
   * @param x x
   * @param y y
   */
  public void draw(int x, int y){
    draw(x, y, textureRectangle, 1.0f, 1.0f, 0.0f);
  }

  /**
   * Draws the texture.
   *
   * @param x x
   * @param y y
   * @param scaleHorizontal horizontal stretch factor
   * @param scaleVertical vertical stretch factor
   * @param angle angle (0..360)
   */
  public void draw(int x, int y, float scaleHorizontal, float scaleVertical, float angle){
    draw(x, y, textureRectangle, scaleHorizontal, scaleVertical, angle);
  }

  /**
   * Draws the texture.
   *
   * @param x x
   * @param y y
   * @param centerX rotation center
   * @param centerY rotation center
   * @param scaleHorizontal horizontal stretch factor
   * @param scaleVertical vertical stretch factor
   * @param angle angle (0..360)
   */
  public void draw(int x, int y, int centerX, int centerY, float scaleHorizontal, float scaleVertical, float angle){
    draw(x, y, textureRectangle, centerX, centerY, scaleHorizontal, scaleVertical, angle);
  }

  /**
   * Draws the texture.
   *
   * @param x x
   * @param y y
   * @param part the part of the image to draw
   * @param scaleHorizontal horizontal stretch factor
   * @param scaleVertical vertical stretch factor
   * @param angle angle (0..360)
   */
  public void draw(int x, int y, Rectangle part, float scaleHorizontal, float scaleVertical, float angle){
    int centerX = part.width / 2;
    int centerY = part.height / 2;

    draw(x, y, part, centerX, centerY, scaleHorizontal, scaleVertical, angle);
  }

  /**
   * Draws the texture.
   *
   * @param x x
   * @param y y
   * @param part the part of the image to draw
   * @param centerX rotation center
   * @param centerY rotation center
   * @param scaleHorizontal horizontal stretch factor
   * @param scaleVertical vertical stretch factor
   * @param angle angle (0..360)
   */
  public void draw(int x, int y, Rectangle part, int centerX, int centerY, float scaleHorizontal, float scaleVertical, float angle){
    centerX += x;
    centerY += y;

    // rotation
    GL.glMatrixMode(GL.GL_MODELVIEW);

    GL.glTranslatef(centerX, centerY, 0);
    GL.glRotatef(angle, 0.0f, 0.0f, -1.0f);
    GL.glScalef(scaleHorizontal, scaleVertical, 1.0f);
    GL.glTranslatef(-centerX, -centerY, 0);

    // draw the texture
    float textureLeft = (float) part.x / textureInternal.width;
    float textureRight = (float) (part.x + part.width) / textureInternal.width;
    float textureTop = (float) part.y / textureInternal.width;
    float textureBottom = (float) (part.y + part.height) / textureInternal.width;

    GL.glBindTexture(GL.GL_TEXTURE_2D, textureInternal.nr);
    GL.glBegin(GL.GL_QUADS);
    {
      GL.glTexCoord2f(textureLeft, textureTop);
      GL.glVertex2i(x, y);
      GL.glTexCoord2f(textureRight, textureTop);
      GL.glVertex2i(x + part.width, y);
      GL.glTexCoord2f(textureRight, textureBottom);
      GL.glVertex2i(x + part.width, y + part.height);
      GL.glTexCoord2f(textureLeft, textureBottom);
      GL.glVertex2i(x, y + part.height);
    }
    GL.glEnd();

    GL.glLoadIdentity();
  }

  /**
   * Draws the texture with horizontal quads.
   * The textureX coordinates are mapped to screenX coordinates (relative to x).
   *
   * @param x x
   * @param y y
   * @param textureX texture coordinates
   * @param screenX screen coordinates (relative to x)
   */
  public void drawHorizontalQuads(int x, int y, int textureX[], int screenX[]){
    GL.glBindTexture(GL.GL_TEXTURE_2D, textureInternal.nr);
    GL.glBegin(GL.GL_QUAD_STRIP);
    {
      for(int part_nr = 0; part_nr < textureX.length; part_nr++){
        float textureXf = (float) textureX[part_nr] / textureInternal.width;

        GL.glTexCoord2f(textureXf, 0.0f);
        GL.glVertex2i(x + screenX[part_nr], y);
        GL.glTexCoord2f(textureXf, 1.0f);
        GL.glVertex2i(x + screenX[part_nr], y + textureInternal.height);
      }
    }
    GL.glEnd();
  }
}
